__plugin_name__ = "PyDevTools"
__plugin_version__ = "1.1"
__plugin_mainclass__ = "PyDevTools"

import traceback,os,random,sys
import org.bukkit.entity
from org.bukkit.inventory import ItemStack
from inspect import getargspec

from java.lang import Class
from java.io import File
from java.sql import DriverManager

# boilerplate
import org.bukkit as bukkit
from java.util.logging import Level

server = bukkit.Bukkit.getServer()

log = server.getLogger()
CHAT_PREFIX = "[PyDevTools] "
def info(*text):
    log.log(Level.INFO,CHAT_PREFIX+" ".join(map(unicode,text)))
def severe(*text):
    log.log(Level.SEVERE,CHAT_PREFIX+" ".join(map(unicode,text)))
def msg(player,*text):
    player.sendMessage(CHAT_PREFIX+" ".join(map(unicode,text)))
# end boilerplate

class DBConnector(object):
    def __init__(self):
        self.db = None
        from java.lang import Class

        Class.forName("org.sqlite.JDBC")
        self.statements = {}

    def connect(self, path):
        try:
            from java.sql import DriverManager
            self.db = DriverManager.getConnection("jdbc:sqlite:%s"%path);
            self.db.setAutoCommit(False)
        except:
            pass

    def disconnect(self):
        if self.db:
            self.db.close()
            self.db = None

    def sql(self, sql):
        if self.db:
            statement = self.db.createStatement()
            statement.executeUpdate(sql)
            self.db.commit()

    def query(self, sql):
        statement = self.db.createStatement()
        if self.db:
            rs = statement.executeQuery(sql)
            return rs

    def saveToDB(self, table, obj, condition = ""):
        statement = None

        if type(obj) == dict and not condition:
            rows = len(obj.values()[0])
            columns = obj.keys()
            sql = "insert into %s (%s) values (%s) %s;"%(table,",".join(columns),",".join(["?" for i in range(len(columns))]),condition)
            print sql
            statement = self.db.prepareStatement(sql)
            setters = {int:statement.setInt,float:statement.setFloat,str:statement.setString,unicode:statement.setString}
            for i in range(rows):
                for j,c in enumerate(columns):
                    print "trying to set",i,j,c
                    value = obj[c][i]
                    print value
                    setters[type(value)](j+1,value)
                statement.addBatch()
        elif type(obj) == dict:
            sql = "update %s set %s where %s;"%(table,",".join(["%s = %s"%(a,b) for a,b in obj.items()]),condition)
            statement = self.db.prepareStatement(sql)
        elif type(obj) == list and not condition:
            sql = "insert into %s values (%s);"%(table,",".join(["?" for x in obj]),condition)
            statement = self.db.prepareStatement(sql)
            setters = {int:statement.setInt,float:statement.setFloat,str:statement.setString,unicode:statement.setString}
            for i,item in enumerate(obj):
                setters[type(item)](i+1,item)
            statement.addBatch()

        if statement:
            statement.executeBatch()
            self.db.commit()

class TestClass(Listener):
    @CommandHandler(usage="/<command> to get a reply",desc="bar")
    def test(self, sender, args):
        print "Hello from %s, %s!"%(self,sender.getName())
        return True

class PyDevTools(PythonPlugin):
    def __init__(self):
        self.macros = {}

        self.bound = {}

        try:
            self.db = DBConnector()
        except:
            info("No SQLite driver found")
            self.db = None

        self.history = {}
        self.cache = {}
        self.verbose = False
        self.interpreters = []
        self.namespaces = {0:{}}
        self.namespaces[0].update(globals())
        self.namespaces[0].update(locals())
        self.namespaces[0]["plugin"] = self
        self.namespaces[1] = self.namespaces[0].copy()

        self.effects = {}
        self.test = TestClass()

    @CommandHandler
    def test2(self, sender, args):
        print "Hello from %s, %s!"%(self,sender.getName())
        return True

    def onEnable(self):
        self.load()
        info("%s v%s enabled"%(__plugin_name__,__plugin_version__))

    def load(self):
        scripts = {}
        for image in File(self.dataFolder,"scripts").listFiles():
            name,ext = os.path.splitext(image.getName())
            if ext == ".py":
                f = open(image.getCanonicalPath())
                scripts[image.getName()] = f.read()
                f.close()
        if len(scripts):
            info("Loaded %s script%s"%(len(scripts),"s" if len(scripts) != 1 else ""))
        self.macros.update(scripts)

    def onDisable(self):
        info("%s v%s disabled"%(__plugin_name__,__plugin_version__))

    @EventHandler("player.PlayerChatEvent","highest")
    def interceptPlayerChat(self, event):
        if self.captureChat(event.getPlayer(),event.getMessage()):
            event.setCancelled(True)

    @EventHandler("server.ServerCommandEvent","highest")
    def interceptServerChat(self, event):
        if self.captureChat(event.getSender(),event.getCommand()):
            event.setCommand("pydummy")

    def captureChat(self, sender, message):
        capture = False
        if sender in self.cache:
            self.cacheLine(sender, message)
            capture = True
        elif sender in self.interpreters:
            capture = True
            if message == "stop":
                capture = False
                message = "exit"
            if message == "exit":
                self.interpreters.remove(sender)
                msg(sender,"Left interactive mode")
            else:
                self.prepare(sender,message)

        return capture

    @CommandHandler("pydummy")
    def dummyCommand(self, sender, args):
        return True

    @CommandHandler("py", usage="/<command> [statement|macro_name]",
                desc="Execute python code or enter interactive interpreter")
    def mainCommand(self,sender,args):
        if not sender.hasPermission("pydev.execute"):
            msg(sender,"You do not have the permission to execute code")
        else:
            if not args and not (sender in self.cache and sender not in self.interpreters):
                separate_namespace = False
                self.interpreters.append(sender)
                if sender not in self.namespaces and separate_namespace:
                    self.namespaces[sender] = self.namespaces[1].copy()
                msg(sender,"Now in interactive mode")
            else:
                if len(args) == 1 and args[0] == "verbose":
                    self.verbose = not self.verbose
                    msg(sender,"Set verbose execution mode to %s"%self.verbose)
                elif len(args) == 1 and args[0] == "reload":
                    self.load()
                elif len(args) == 1 and args[0] in self.macros:
                    msg(sender,"[Macro] "+args[0])
                    self.execute(sender,self.macros[args[0]])
                elif len(args) == 2 and args[0] == "save" and args[1] == "history":
                    f = open(File(self.dataFolder,"history_%s.txt"%sender.getName()).getCanonicalPath(),"w")
                    f.write("\n".join(self.history.get(sender)))
                    f.close()
                    msg(sender,"Saved history for user '%s'"%sender.getName())
                elif len(args) == 3 and args[0] == "save" and args[1] in self.macros:
                    macro = args[1]
                    filename = args[2]
                    filename += "" if filename.endswith(".py") else ".py"
                    f = open(File(self.dataFolder,"scripts/%s"%filename).getCanonicalPath(),"w")
                    f.write(self.macros[macro])
                    f.close()
                    self.macros[filename+".py"] = self.macros[macro]
                    msg(sender,"Saved macro '%s' as '%s'"%(macro,filename))
                else:
                    self.prepare(sender, " ".join(args))
        return True

    @CommandHandler("pymacro", usage="/<command> macro_name statement",
                desc="save python code as macro to execute with /exec")
    def macroCommand(self,sender,args):
        if not sender.hasPermission("pydev.execute"):
            msg(sender,"You do not have the permission to execute code")
        else:
            if len(args) < 2:
                msg(sender,"No statement given to save as macro")
            else:
                self.saveMacro(sender, args[0]," ".join(args[1:]))
        return True

    def saveMacro(self, sender, name, statement):
        self.history.setdefault(sender,[]).append(statement)
        self.macros[name] = statement
        msg(sender,"[Success] Defined macro '%s'"%name)

    @CommandHandler("bind", usage="/<command> macro_name",
                    desc="bind saved macro to current tool")
    def bind(self,sender,args):
        self.bound.setdefault(sender, {})
        if not sender.hasPermission("pydev.execute"):
            msg(sender,"You do not have the permission to execute code")
        else:
            material = sender.getItemInHand().getType()
            if not args and material in self.bound[sender]:
                del self.bound[sender][material]
            else:
                self.bound[sender][material] = args[0]
        return True

    def prepare(self, sender, statement = ""):
        self.namespaces.get(sender,self.namespaces[0])["self"] = sender
        flushed_cache = False

        if statement.endswith("..."):
            if not sender in self.cache:
                self.cacheStart(sender)
            self.cacheLine(sender,statement[:-3])
            statement = ""
        elif not statement and sender in self.cache:
            statement = self.cacheFlush(sender)
            flushed_cache = True

        if statement:
            if sender in self.interpreters or flushed_cache:
                sender.sendMessage(">>> "+ (not flushed_cache)*statement)
            self.execute(sender, statement)

    def execute(self, sender, statement):
        self.macros["last"] = statement
        self.history.setdefault(sender,[]).append(statement)
        code = compile(statement,"<string>","exec")
        try:
            exec code in self.namespaces.get(sender,self.namespaces[0])

            if sender not in self.interpreters:
                msg(sender,"[Success] "+ self.verbose*statement)
        except:
            msg(sender,"[Failure] "+ self.verbose*statement)
            log.log(Level.SEVERE,traceback.format_exc())

    def cacheStart(self, sender):
        self.cache[sender] = []
        if sender not in self.interpreters:
            msg(sender,"Started multiline statement, leading '.' indents, '.' to execute, 'macro name' to save as 'name'")

    def cacheLine(self, sender, message):
        if message.startswith("macro") and len(message.split()) == 2:
            name = message.split()[1]
            self.saveMacro(sender, name,self.cacheFlush(sender))
        elif not message.replace("."," ").strip():
            self.prepare(sender)
        else:
            indendation = ""
            while message.startswith("."):
                message = message[1:]
                indendation += "    "
            if not self.cache[sender]:
                sender.sendMessage(">>> "+indendation+message)
            else:
                sender.sendMessage("*** "+indendation+message)
            self.cache[sender].append(indendation + message)

    def cacheFlush(self, sender):
        statement = "\n".join(self.cache[sender])
        del self.cache[sender]
        return statement

    @EventHandler("player.PlayerInteractEvent","highest")
    def checkBindingBlock(self,event):
        event.setCancelled(self.checkBinding(event.getPlayer(),event.getClickedBlock(),None))

    @EventHandler("player.PlayerInteractEntityEvent","highest")
    def checkBindingEntity(self, event):
        event.setCancelled(self.checkBinding(event.getPlayer(),None,event.getRightClicked()))

    def checkBinding(self, player, block, entity):
        macro = None
        try:
            macro = self.bound.setdefault(player,{}).get(player.getItemInHand().getType())
        except:
            pass
        if macro in self.macros:
            ns = self.namespaces.get(event.getPlayer(),self.namespaces[0])
            ns["block"] = block
            ns["entity"] = entity
            self.execute(player,self.macros[macro])
            return True
        else:
            return False

@CommandHandler("effect",usage="/<command> [effect] [additional arguments] [repetitions]",desc="play effect")
def bindeffect(sender, args):
    if not sender.hasPermission("pytest.execute"):
        msg(sender,"You do not have the permission to execute code")
        return True

    if len(args) == 0:
        msg(sender,"too few arguments")
        return True

    try:
        # effect, data, repetitions
        pyplugin.effects["current"] = (eval("bukkit.Effect."+args[0].upper()),args[1] if len(args) > 1 else "0",args[2] if len(args) > 2 else "1")
        sender.sendMessage("Effect set to %s"%args[0])
    except:
        msg(sender,"error finding specified effect")
    return True

@EventHandler("entity.ProjectileHitEvent","monitor")
def playeffect(event):
    entity = event.getEntity()
    if isinstance(entity,org.bukkit.entity.Snowball):
        if "current" in pyplugin.effects:
            effect,data_range,repetitions = pyplugin.effects["current"]
            for i in range(int(repetitions)):
                data = 0
                if "-" in data_range:
                    data = random.randint(*map(int,data_range.split("-")))
                else:
                    data = int(data_range)
                entity.getWorld().playEffect(entity.getLocation(),effect,data)